@@ -1,7 +1,7 @@
 import asyncio
 import json
 import logging
-from abc import ABC
+from abc import ABC, abstractmethod
 from asyncio import Lock
 from datetime import datetime
 from functools import partial
@@ -42,11 +42,27 @@ from cilroy.controller.parameters import (
     ScrapAfterParameter,
     ScrapBeforeParameter,
     ScrapLimitParameter,
+    FeedLengthParameter,
+)
+from cilroy.controller.params import (
+    OfflineParams,
+    OnlineParams,
+    Params,
+    FaceParams,
+    ModuleParams,
+    FeedParams,
+)
+from cilroy.controller.state import (
+    OfflineState,
+    OnlineState,
+    State,
+    FaceState,
+    ModuleState,
+    FeedState,
 )
-from cilroy.controller.params import OfflineParams, OnlineParams, Params
-from cilroy.controller.state import OfflineState, OnlineState, State
 from cilroy.messages import Status
 from cilroy.metadata import Metadata
+from cilroy.post import PostData
 from cilroy.posting import PostScheduler
 from cilroy.scoring import ScoreScheduler
 from cilroy.status import TrainingStatus
@@ -57,12 +73,26 @@ logger = logging.getLogger(__name__)
 
 class CilroyControllerBase(Configurable[State], ABC):
     @staticmethod
-    async def _build_face_service(params: Params) -> FaceService:
-        return FaceService(Channel(params.face_host, params.face_port))
+    async def _build_face_service(params: FaceParams) -> FaceService:
+        return FaceService(Channel(params.host, params.port))
+
+    @classmethod
+    async def _build_face_state(cls, params: FaceParams) -> FaceState:
+        return FaceState(
+            service=await cls._build_face_service(params),
+        )
 
     @staticmethod
-    async def _build_module_service(params: Params) -> ModuleService:
-        return ModuleService(Channel(params.module_host, params.module_port))
+    async def _build_module_service(params: ModuleParams) -> ModuleService:
+        return ModuleService(Channel(params.host, params.port))
+
+    async def _build_module_state(self, params: ModuleParams) -> ModuleState:
+        return ModuleState(
+            service=await self._build_module_service(params),
+            archived_metrics=[],
+            metrics_stream=await Observable.build(),
+            metrics_task=asyncio.create_task(self._watch_metrics_task()),
+        )
 
     @staticmethod
     async def _build_offline_state(params: OfflineParams) -> OfflineState:
@@ -116,15 +146,24 @@ class CilroyControllerBase(Configurable[State], ABC):
     async def _build_training_status() -> Observable[TrainingStatus]:
         return await Observable.build(TrainingStatus.IDLE)
 
+    @staticmethod
+    async def _build_feed_state(params: FeedParams) -> FeedState:
+        return FeedState(
+            feed=[],
+            length=params.length,
+            stream=await Observable.build(),
+        )
+
     async def _build_default_state(self) -> State:
         params = Params(**self._kwargs)
         return State(
-            face_service=await self._build_face_service(params),
-            module_service=await self._build_module_service(params),
+            face=await self._build_face_state(params.face),
+            module=await self._build_module_state(params.module),
             offline=await self._build_offline_state(params.offline),
             online=await self._build_online_state(params.online),
             training_task=None,
             training_status=await self._build_training_status(),
+            feed=await self._build_feed_state(params.feed),
         )
 
     @staticmethod
@@ -134,6 +173,28 @@ class CilroyControllerBase(Configurable[State], ABC):
         with open(directory / "state.json", "w") as f:
             json.dump(state_dict, f)
 
+    @staticmethod
+    async def _save_face_state(state: FaceState, directory: Path) -> None:
+        pass
+
+    @staticmethod
+    async def _save_module_metrics(
+        metrics: List[MetricData], directory: Path
+    ) -> None:
+        serialized = [json.loads(metric.json()) for metric in metrics]
+        with open(directory / "metrics.json", "w") as f:
+            json.dump([metric for metric in serialized], f)
+
+    @classmethod
+    async def _save_module_state(
+        cls, state: ModuleState, directory: Path
+    ) -> None:
+        metrics_directory = directory / "metrics"
+        metrics_directory.mkdir(parents=True, exist_ok=True)
+        await cls._save_module_metrics(
+            state.archived_metrics, metrics_directory
+        )
+
     @staticmethod
     async def _create_offline_state_dict(
         state: OfflineState,
@@ -217,8 +278,25 @@ class CilroyControllerBase(Configurable[State], ABC):
         state_dict = await cls._create_online_state_dict(state)
         await cls._save_state_dict(directory, state_dict)
 
+    @classmethod
+    async def _save_feed_state(cls, state: FeedState, directory: Path) -> None:
+        serialized_feed = [json.loads(post.json()) for post in state.feed]
+        state_dict = {
+            "feed": serialized_feed,
+            "length": state.length,
+        }
+        await cls._save_state_dict(directory, state_dict)
+
     @classmethod
     async def _save_state(cls, state: State, directory: Path) -> None:
+        face_directory = directory / "face"
+        face_directory.mkdir(parents=True, exist_ok=True)
+        await cls._save_face_state(state.face, face_directory)
+
+        module_directory = directory / "module"
+        module_directory.mkdir(parents=True, exist_ok=True)
+        await cls._save_module_state(state.module, module_directory)
+
         offline_directory = directory / "offline"
         offline_directory.mkdir(parents=True, exist_ok=True)
         await cls._save_offline_state(state.offline, offline_directory)
@@ -227,11 +305,39 @@ class CilroyControllerBase(Configurable[State], ABC):
         online_directory.mkdir(parents=True, exist_ok=True)
         await cls._save_online_state(state.online, online_directory)
 
+        feed_directory = directory / "feed"
+        feed_directory.mkdir(parents=True, exist_ok=True)
+        await cls._save_feed_state(state.feed, feed_directory)
+
     @staticmethod
     async def _load_state_dict(directory: Path) -> Dict[str, Any]:
         with open(directory / "state.json", "r") as f:
             return json.load(f)
 
+    @classmethod
+    async def _load_saved_face_state(
+        cls, directory: Path, params: FaceParams
+    ) -> FaceState:
+        return FaceState(service=await cls._build_face_service(params))
+
+    @staticmethod
+    async def _load_module_metrics(directory: Path) -> List[MetricData]:
+        with open(directory / "metrics.json", "r") as f:
+            serialized = json.load(f)
+        return [MetricData.parse_obj(metric) for metric in serialized]
+
+    async def _load_saved_module_state(
+        self, directory: Path, params: ModuleParams
+    ) -> ModuleState:
+        return ModuleState(
+            service=await self._build_module_service(params),
+            archived_metrics=await self._load_module_metrics(
+                directory / "metrics"
+            ),
+            metrics_stream=await Observable.build(),
+            metrics_task=asyncio.create_task(self._watch_metrics_task()),
+        )
+
     @classmethod
     async def _load_saved_offline_state(
         cls, directory: Path, params: OfflineParams
@@ -310,11 +416,28 @@ class CilroyControllerBase(Configurable[State], ABC):
             lock=Lock(),
         )
 
+    @classmethod
+    async def _load_saved_feed_state(
+        cls, directory: Path, params: FeedParams
+    ) -> FeedState:
+        state_dict = await cls._load_state_dict(directory)
+        return FeedState(
+            feed=[
+                PostData.parse_obj(post) for post in state_dict.get("feed", [])
+            ],
+            length=state_dict.get("length", params.length),
+            stream=await Observable.build(),
+        )
+
     async def _load_saved_state(self, directory: Path) -> State:
         params = Params(**self._kwargs)
         return State(
-            face_service=await self._build_face_service(params),
-            module_service=await self._build_module_service(params),
+            face=await self._load_saved_face_state(
+                directory / "face", params.face
+            ),
+            module=await self._load_saved_module_state(
+                directory / "module", params.module
+            ),
             offline=await self._load_saved_offline_state(
                 directory / "offline", params.offline
             ),
@@ -323,8 +446,19 @@ class CilroyControllerBase(Configurable[State], ABC):
             ),
             training_task=None,
             training_status=await self._build_training_status(),
+            feed=await self._load_saved_feed_state(
+                directory / "feed", params.feed
+            ),
         )
 
+    async def cleanup(self) -> None:
+        async with self.state.write_lock() as state:
+            state.module.metrics_task.cancel()
+            try:
+                await state.module.metrics_task
+            except asyncio.CancelledError:
+                pass
+
     @classproperty
     def parameters(cls) -> Set[Parameter]:
         return {
@@ -337,79 +471,84 @@ class CilroyControllerBase(Configurable[State], ABC):
             ScoreSchedulerParameter(),
             OnlineIterationsParameter(),
             OnlineBatchSizeParameter(),
+            FeedLengthParameter(),
         }
 
+    @abstractmethod
+    async def _watch_metrics_task(self) -> None:
+        pass
 
-class CilroyControllerDelegatedBase(CilroyControllerBase):
+
+class CilroyControllerDelegatedBase(CilroyControllerBase, ABC):
     async def get_face_metadata(self) -> Metadata:
         async with self.state.read_lock() as state:
-            metadata = await state.face_service.get_metadata()
+            metadata = await state.face.service.get_metadata()
             return Metadata(key=metadata.key, description=metadata.description)
 
     async def get_module_metadata(self) -> Metadata:
         async with self.state.read_lock() as state:
-            metadata = await state.module_service.get_metadata()
+            metadata = await state.module.service.get_metadata()
             return Metadata(key=metadata.key, description=metadata.description)
 
     async def get_face_post_schema(self) -> JSONSchema:
         async with self.state.read_lock() as state:
-            return JSONSchema(**await state.face_service.get_post_schema())
+            return JSONSchema(**await state.face.service.get_post_schema())
 
     async def get_module_post_schema(self) -> JSONSchema:
         async with self.state.read_lock() as state:
-            return JSONSchema(**await state.module_service.get_post_schema())
+            return JSONSchema(**await state.module.service.get_post_schema())
 
     async def get_face_status(self) -> Status:
         async with self.state.read_lock() as state:
-            status = await state.face_service.get_status()
+            status = await state.face.service.get_status()
         return Status(status)
 
     async def watch_face_status(self) -> AsyncIterable[Status]:
         async with self.state.read_lock() as state:
-            service = state.face_service
+            service = state.face.service
         async for status in service.watch_status():
             yield Status(status)
 
     async def get_module_status(self) -> Status:
         async with self.state.read_lock() as state:
-            status = await state.module_service.get_status()
+            status = await state.module.service.get_status()
         return Status(status)
 
     async def watch_module_status(self) -> AsyncIterable[Status]:
         async with self.state.read_lock() as state:
-            service = state.module_service
+            service = state.module.service
         async for status in service.watch_status():
             yield Status(status)
 
     async def get_face_config_schema(self) -> JSONSchema:
         async with self.state.read_lock() as state:
-            return JSONSchema(**await state.face_service.get_config_schema())
+            return JSONSchema(**await state.face.service.get_config_schema())
 
     async def get_face_config(self) -> Dict[str, Any]:
         async with self.state.read_lock() as state:
-            return await state.face_service.get_config()
+            return await state.face.service.get_config()
 
     async def watch_face_config(self) -> AsyncIterable[Dict[str, Any]]:
         async with self.state.read_lock() as state:
-            service = state.face_service
+            service = state.face.service
         async for config in service.watch_config():
             yield config
 
     async def set_face_config(self, config: Dict[str, Any]) -> Dict[str, Any]:
         async with self.state.write_lock() as state:
-            return await state.face_service.set_config(config)
+            return await state.face.service.set_config(config)
 
     async def get_module_config_schema(self) -> JSONSchema:
         async with self.state.read_lock() as state:
-            return JSONSchema(**await state.module_service.get_config_schema())
+            return JSONSchema(**await state.module.service.get_config_schema())
 
     async def get_module_config(self) -> Dict[str, Any]:
         async with self.state.read_lock() as state:
-            return await state.module_service.get_config()
+            return await state.module.service.get_config()
 
     async def watch_module_config(self) -> AsyncIterable[Dict[str, Any]]:
         async with self.state.read_lock() as state:
-            service = state.module_service
+            service = state.module.service
         async for config in service.watch_config():
             yield config
 
@@ -417,7 +556,7 @@ class CilroyControllerDelegatedBase(CilroyControllerBase):
         self, config: Dict[str, Any]
     ) -> Dict[str, Any]:
         async with self.state.write_lock() as state:
-            return await state.module_service.set_config(config)
+            return await state.module.service.set_config(config)
 
 
 class CilroyController(CilroyControllerDelegatedBase):
@@ -436,7 +575,7 @@ class CilroyController(CilroyControllerDelegatedBase):
             async with self.state.read_lock() as state:
                 max_epochs = state.offline.max_epochs
                 batch_size = state.offline.batch_size
-                service = state.module_service
+                service = state.module.service
 
             if max_epochs is not None and epoch >= max_epochs:
                 break
@@ -463,7 +602,7 @@ class CilroyController(CilroyControllerDelegatedBase):
                 yield post, score
 
         state = await self.state.value.fetch()
-        posts = state.face_service.scrap(
+        posts = state.face.service.scrap(
             state.offline.scrap_limit,
             state.offline.scrap_before,
             state.offline.scrap_after,
@@ -488,8 +627,8 @@ class CilroyController(CilroyControllerDelegatedBase):
 
     async def _train_online_post_loop(self) -> None:
         async with self.state.read_lock() as state:
-            module_service = state.module_service
-            face_service = state.face_service
+            module_service = state.module.service
+            face_service = state.face.service
             post_scheduler = state.online.post_scheduler
             lock = state.online.lock
 
@@ -498,9 +637,20 @@ class CilroyController(CilroyControllerDelegatedBase):
                 logger.info("Online training: Creating new post...")
 
                 module_post_id, post = await anext(module_service.generate())
-                face_post_id = await face_service.post(post)
+                face_post_id, post_url = await face_service.post(post)
+
+                post_data = PostData(
+                    id=face_post_id,
+                    url=post_url,
+                    content=post,
+                    created_at=datetime.utcnow(),
+                )
+
                 async with self.state.write_lock() as state:
                     state.online.ids_cache[face_post_id] = module_post_id
+                    state.feed.feed = state.feed.feed + [post_data]
+                    state.feed.feed = state.feed.feed[-state.feed.length :]
+                    await state.feed.stream.set(post_data)
 
                 logger.info(
                     f"Online training: "
@@ -517,7 +667,7 @@ class CilroyController(CilroyControllerDelegatedBase):
             async with self.state.read_lock() as state:
                 iterations = state.online.iterations
                 batch_size = state.online.batch_size
-                service = state.module_service
+                service = state.module.service
 
             if iteration >= iterations:
                 break
@@ -529,7 +679,7 @@ class CilroyController(CilroyControllerDelegatedBase):
 
     async def _train_online_score_loop(self) -> None:
         async with self.state.read_lock() as state:
-            face_service = state.face_service
+            face_service = state.face.service
             score_scheduler = state.online.score_scheduler
             lock = state.online.lock
 
@@ -601,9 +751,53 @@ class CilroyController(CilroyControllerDelegatedBase):
 
     async def get_module_metrics_config(self) -> List[MetricConfig]:
         async with self.state.read_lock() as state:
-            return await state.module_service.get_metrics_config()
+            return await state.module.service.get_metrics_config()
+
+    async def get_module_metrics(self) -> List[MetricData]:
+        async with self.state.read_lock() as state:
+            return state.module.archived_metrics
+
+    async def _watch_metrics_task(self) -> None:
+        async with self.state.read_lock() as state:
+            service: ModuleService = state.module.service
+
+        while True:
+            try:
+                async for metric in service.watch_metrics():
+                    async with self.state.write_lock() as state:
+                        state.module.archived_metrics.append(metric)
+                        await state.module.metrics_stream.set(metric)
+            except asyncio.CancelledError:
+                raise
+            except Exception as e:
+                logger.warning("Metrics stream failed.", exc_info=e)
+                await asyncio.sleep(1)
 
     async def watch_module_metrics(self) -> AsyncIterator[MetricData]:
-        state = await self.state.value.fetch()
-        async for metric in state.module_service.watch_metrics():
+        async with self.state.read_lock() as state:
+            metrics = state.module.metrics_stream
+        async for metric in metrics.subscribe():
             yield metric
+
+    async def reset_face(self) -> None:
+        async with self.state.write_lock() as state:
+            await state.face.service.reset()
+
+    async def reset_module(self) -> None:
+        await self.stop_training()
+        async with self.state.write_lock() as state:
+            await state.module.service.reset()
+            state.module.archived_metrics = []
+            state.offline.posts_cache.clear()
+            state.online.ids_cache.clear()
+            state.feed.feed = []
+
+    async def get_feed(self) -> List[PostData]:
+        async with self.state.read_lock() as state:
+            return state.feed.feed
+
+    async def watch_feed(self) -> AsyncIterator[PostData]:
+        async with self.state.read_lock() as state:
+            feed = state.feed.stream
+        async for post in feed.subscribe():
+            yield post
