@@ -1,6 +1,7 @@
 import json
-from typing import AsyncIterator, Dict
+from typing import AsyncIterator, Dict, Tuple
 
+import aiostream
 import grpclib
 from betterproto.grpc.grpclib_server import ServiceBase
 from grpclib.server import Stream
@@ -152,11 +153,46 @@ class CilroyServiceBase(ServiceBase):
     ) -> "GetModuleMetricsConfigResponse":
         raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)
 
+    async def get_module_metrics(
+        self, get_module_metrics_request: "GetModuleMetricsRequest"
+    ) -> "GetModuleMetricsResponse":
+        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)
+
     async def watch_module_metrics(
         self, watch_module_metrics_request: "WatchModuleMetricsRequest"
     ) -> AsyncIterator["WatchModuleMetricsResponse"]:
         raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)
 
+    async def watch_all(
+        self, watch_all_request: "WatchAllRequest"
+    ) -> AsyncIterator["WatchAllResponse"]:
+        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)
+
+    async def reset_controller(
+        self, reset_controller_request: "ResetControllerRequest"
+    ) -> "ResetControllerResponse":
+        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)
+
+    async def reset_face(
+        self, reset_face_request: "ResetFaceRequest"
+    ) -> "ResetFaceResponse":
+        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)
+
+    async def reset_module(
+        self, reset_module_request: "ResetModuleRequest"
+    ) -> "ResetModuleResponse":
+        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)
+
+    async def get_feed(
+        self, get_feed_request: "GetFeedRequest"
+    ) -> "GetFeedResponse":
+        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)
+
+    async def watch_feed(
+        self, watch_feed_request: "WatchFeedRequest"
+    ) -> AsyncIterator["WatchFeedResponse"]:
+        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)
+
     async def __rpc_get_face_metadata(
         self,
         stream: "grpclib.server.Stream[GetFaceMetadataRequest, GetFaceMetadataResponse]",
@@ -402,6 +438,14 @@ class CilroyServiceBase(ServiceBase):
         response = await self.get_module_metrics_config(request)
         await stream.send_message(response)
 
+    async def __rpc_get_module_metrics(
+        self,
+        stream: "grpclib.server.Stream[GetModuleMetricsRequest, GetModuleMetricsResponse]",
+    ) -> None:
+        request = await stream.recv_message()
+        response = await self.get_module_metrics(request)
+        await stream.send_message(response)
+
     async def __rpc_watch_module_metrics(
         self,
         stream: "grpclib.server.Stream[WatchModuleMetricsRequest, WatchModuleMetricsResponse]",
@@ -413,6 +457,59 @@ class CilroyServiceBase(ServiceBase):
             request,
         )
 
+    async def __rpc_watch_all(
+        self,
+        stream: "grpclib.server.Stream[WatchAllRequest, WatchAllResponse]",
+    ) -> None:
+        request = await stream.recv_message()
+        await self._call_rpc_handler_server_stream(
+            self.watch_all,
+            stream,
+            request,
+        )
+
+    async def __rpc_reset_controller(
+        self,
+        stream: "grpclib.server.Stream[ResetControllerRequest, ResetControllerResponse]",
+    ) -> None:
+        request = await stream.recv_message()
+        response = await self.reset_controller(request)
+        await stream.send_message(response)
+
+    async def __rpc_reset_face(
+        self,
+        stream: "grpclib.server.Stream[ResetFaceRequest, ResetFaceResponse]",
+    ) -> None:
+        request = await stream.recv_message()
+        response = await self.reset_face(request)
+        await stream.send_message(response)
+
+    async def __rpc_reset_module(
+        self,
+        stream: "grpclib.server.Stream[ResetModuleRequest, ResetModuleResponse]",
+    ) -> None:
+        request = await stream.recv_message()
+        response = await self.reset_module(request)
+        await stream.send_message(response)
+
+    async def __rpc_get_feed(
+        self, stream: "grpclib.server.Stream[GetFeedRequest, GetFeedResponse]"
+    ) -> None:
+        request = await stream.recv_message()
+        response = await self.get_feed(request)
+        await stream.send_message(response)
+
+    async def __rpc_watch_feed(
+        self,
+        stream: "grpclib.server.Stream[WatchFeedRequest, WatchFeedResponse]",
+    ) -> None:
+        request = await stream.recv_message()
+        await self._call_rpc_handler_server_stream(
+            self.watch_feed,
+            stream,
+            request,
+        )
+
     def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
         return {
             "/kilroy.cilroy.v1alpha.CilroyService/GetFaceMetadata": grpclib.const.Handler(
@@ -583,12 +680,54 @@ class CilroyServiceBase(ServiceBase):
                 GetModuleMetricsConfigRequest,
                 GetModuleMetricsConfigResponse,
             ),
+            "/kilroy.cilroy.v1alpha.CilroyService/GetModuleMetrics": grpclib.const.Handler(
+                self.__rpc_get_module_metrics,
+                grpclib.const.Cardinality.UNARY_UNARY,
+                GetModuleMetricsRequest,
+                GetModuleMetricsResponse,
+            ),
             "/kilroy.cilroy.v1alpha.CilroyService/WatchModuleMetrics": grpclib.const.Handler(
                 self.__rpc_watch_module_metrics,
                 grpclib.const.Cardinality.UNARY_STREAM,
                 WatchModuleMetricsRequest,
                 WatchModuleMetricsResponse,
             ),
+            "/kilroy.cilroy.v1alpha.CilroyService/WatchAll": grpclib.const.Handler(
+                self.__rpc_watch_all,
+                grpclib.const.Cardinality.UNARY_STREAM,
+                WatchAllRequest,
+                WatchAllResponse,
+            ),
+            "/kilroy.cilroy.v1alpha.CilroyService/ResetController": grpclib.const.Handler(
+                self.__rpc_reset_controller,
+                grpclib.const.Cardinality.UNARY_UNARY,
+                ResetControllerRequest,
+                ResetControllerResponse,
+            ),
+            "/kilroy.cilroy.v1alpha.CilroyService/ResetFace": grpclib.const.Handler(
+                self.__rpc_reset_face,
+                grpclib.const.Cardinality.UNARY_UNARY,
+                ResetFaceRequest,
+                ResetFaceResponse,
+            ),
+            "/kilroy.cilroy.v1alpha.CilroyService/ResetModule": grpclib.const.Handler(
+                self.__rpc_reset_module,
+                grpclib.const.Cardinality.UNARY_UNARY,
+                ResetModuleRequest,
+                ResetModuleResponse,
+            ),
+            "/kilroy.cilroy.v1alpha.CilroyService/GetFeed": grpclib.const.Handler(
+                self.__rpc_get_feed,
+                grpclib.const.Cardinality.UNARY_UNARY,
+                GetFeedRequest,
+                GetFeedResponse,
+            ),
+            "/kilroy.cilroy.v1alpha.CilroyService/WatchFeed": grpclib.const.Handler(
+                self.__rpc_watch_feed,
+                grpclib.const.Cardinality.UNARY_STREAM,
+                WatchFeedRequest,
+                WatchFeedResponse,
+            ),
         }
 
 
@@ -683,7 +822,7 @@ class CilroyService(CilroyServiceBase):
     ) -> AsyncIterator["WatchTrainingStatusResponse"]:
         status = await self._controller.get_training_status()
         async for status in status.subscribe():
-            yield GetTrainingStatusResponse(
+            yield WatchTrainingStatusResponse(
                 status=TrainingStatus(status.value)
             )
 
@@ -815,18 +954,143 @@ class CilroyService(CilroyServiceBase):
                 MetricConfig(
                     id=metric.id,
                     label=metric.label,
+                    group=metric.group,
                     config=json.dumps(metric.config),
                 )
                 for metric in metrics
             ]
         )
 
+    async def get_module_metrics(
+        self, get_module_metrics_request: "GetModuleMetricsRequest"
+    ) -> "GetModuleMetricsResponse":
+        metrics = await self._controller.get_module_metrics()
+        return GetModuleMetricsResponse(
+            metrics=[
+                MetricData(
+                    metric_id=metric.metric_id,
+                    dataset_id=metric.dataset_id,
+                    data=json.dumps(metric.data),
+                )
+                for metric in metrics
+            ]
+        )
+
     async def watch_module_metrics(
         self, watch_module_metrics_request: "WatchModuleMetricsRequest"
     ) -> AsyncIterator["WatchModuleMetricsResponse"]:
         async for metric in self._controller.watch_module_metrics():
             yield WatchModuleMetricsResponse(
-                metric_id=metric.metric_id,
-                dataset_id=metric.dataset_id,
-                data=json.dumps(metric.data),
+                metric=MetricData(
+                    metric_id=metric.metric_id,
+                    dataset_id=metric.dataset_id,
+                    data=json.dumps(metric.data),
+                )
+            )
+
+    async def watch_all(
+        self, watch_all_request: "WatchAllRequest"
+    ) -> AsyncIterator["WatchAllResponse"]:
+        async def convert(
+            method: str, stream: AsyncIterator[betterproto.Message]
+        ) -> AsyncIterator[Tuple[str, str]]:
+            async for msg in stream:
+                yield method, msg.to_json()
+
+        streams = [
+            (
+                "WatchControllerStatus",
+                self.watch_controller_status(WatchControllerStatusRequest()),
+            ),
+            (
+                "WatchFaceStatus",
+                self.watch_face_status(WatchFaceStatusRequest()),
+            ),
+            (
+                "WatchModuleStatus",
+                self.watch_module_status(WatchModuleStatusRequest()),
+            ),
+            (
+                "WatchTrainingStatus",
+                self.watch_training_status(WatchTrainingStatusRequest()),
+            ),
+            (
+                "WatchControllerConfig",
+                self.watch_controller_config(WatchControllerConfigRequest()),
+            ),
+            (
+                "WatchFaceConfig",
+                self.watch_face_config(WatchFaceConfigRequest()),
+            ),
+            (
+                "WatchModuleConfig",
+                self.watch_module_config(WatchModuleConfigRequest()),
+            ),
+            (
+                "WatchModuleMetrics",
+                self.watch_module_metrics(WatchModuleMetricsRequest()),
+            ),
+            (
+                "WatchFeed",
+                self.watch_feed(WatchFeedRequest()),
+            ),
+        ]
+
+        combine = aiostream.stream.merge(
+            *(convert(method, stream) for method, stream in streams)
+        )
+
+        async with combine.stream() as streamer:
+            async for method, message in streamer:
+                yield WatchAllResponse(method=method, message=message)
+
+    async def reset_controller(
+        self, reset_controller_request: "ResetControllerRequest"
+    ) -> "ResetControllerResponse":
+        await self._controller.init()
+        return ResetControllerResponse()
+
+    async def reset_face(
+        self, reset_face_request: "ResetFaceRequest"
+    ) -> "ResetFaceResponse":
+        await self._controller.reset_face()
+        return ResetFaceResponse()
+
+    async def reset_module(
+        self, reset_module_request: "ResetModuleRequest"
+    ) -> "ResetModuleResponse":
+        await self._controller.reset_module()
+        return ResetModuleResponse()
+
+    async def get_feed(
+        self, get_feed_request: "GetFeedRequest"
+    ) -> "GetFeedResponse":
+        feed = await self._controller.get_feed()
+        return GetFeedResponse(
+            posts=[
+                Post(
+                    id=str(post.id),
+                    url=post.url,
+                    content=json.dumps(post.content),
+                    created_at=post.created_at.isoformat().replace(
+                        "+00:00", "Z"
+                    ),
+                )
+                for post in feed
+            ]
+        )
+
+    async def watch_feed(
+        self, watch_feed_request: "WatchFeedRequest"
+    ) -> AsyncIterator["WatchFeedResponse"]:
+        async for post in self._controller.watch_feed():
+            yield WatchFeedResponse(
+                post=Post(
+                    id=str(post.id),
+                    url=post.url,
+                    content=json.dumps(post.content),
+                    created_at=post.created_at.isoformat().replace(
+                        "+00:00", "Z"
+                    ),
+                )
             )
